"""AttriBox provides a descriptor with lazy instantiation of the
underlying object. """
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ..static import AbstractObject
from ..static.zeroton import DELETED
from ..waitaminute import MissingVariable, TypeException, VariableNotNone

try:
  from typing import TYPE_CHECKING
except ImportError:  # pragma: no cover
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any, Self


class AttriBox(AbstractObject):
  """
  AttriBox provides a descriptor with lazy instantiation of the
  underlying object.
  """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Private Variables
  __field_type__ = None

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  GETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def getFieldType(self) -> type:
    """
    Returns the type of the field object.
    """
    if self.__field_type__ is None:
      raise MissingVariable('AttriBox.__field_type__')
    if isinstance(self.__field_type__, type):
      return self.__field_type__
    name, value = '__field_type__', self.__field_type__
    raise TypeException(name, value, type)

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  SETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def setFieldType(self, fieldType: type) -> None:
    """
    Sets the type of the field object.
    """
    if self.__field_type__ is not None:
      raise VariableNotNone('AttriBox.__field_type__', self.__field_type__)
    if not isinstance(fieldType, type):
      raise TypeException('fieldType', fieldType, type)
    self.__field_type__ = fieldType

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  CONSTRUCTORS   # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __init__(self, fieldType: type) -> None:
    """
    Initializes the AttriBox with the type of the object to be returned.
    """
    self.setFieldType(fieldType)

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  Python API   # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  @classmethod
  def __class_getitem__(cls, fieldType: type) -> Self:
    """
    Returns a new AttriBox with the specified field type.
    """
    return cls(fieldType)

  def __call__(self, *args, **kwargs) -> Any:
    """
    This method pretends to be the actual constructor.
    """
    self.__pos_args__ = (*args,)
    self.__key_args__ = {**kwargs, }
    return self

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  DOMAIN SPECIFIC  # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def _createFieldObject(self, ) -> Any:
    """
    Creates the field object using the type specified in the AttriBox.
    This method is used by the 'strongBox' to instantiate the field
    object when it is not already cached.
    """
    fieldType = self.getFieldType()
    ins, own = self.instance, self.owner
    args = self._getPositionalArgs(THIS=ins, OWNER=own)
    kwargs = self._getKeywordArgs(THIS=ins, OWNER=own)
    return fieldType(*args, **kwargs)

  def __instance_get__(self, **kwargs, ) -> Any:
    """
    Attempts to retrieve the value of the field object from the instance.
    If the value is 'None', or if 'getattr' raises 'AttributeError',
    a new 'fieldObject' is created and set on the instance and the
    method recurses to retrieve the value again. This ensures that the
    retrieval does work.
    """
    pvtName = self.getPrivateName()
    try:
      value = getattr(self.instance, pvtName, )
    except AttributeError:
      if kwargs.get('_recursion', False):
        raise RecursionError
      value = self._createFieldObject()
      setattr(self.instance, pvtName, value)
      return self.__instance_get__(_recursion=True)
    else:
      if value is None:
        if kwargs.get('_recursion2', False):
          raise RecursionError
        value = self._createFieldObject()
        setattr(self.instance, pvtName, value)
        return self.__instance_get__(_recursion2=True)
      return value

  def __instance_set__(self, value: Any, **kwargs) -> None:
    """
    Implements type guard.
    """
    fieldType = self.getFieldType()
    if isinstance(value, fieldType):
      pvtName = self.getPrivateName()
      return setattr(self.instance, pvtName, value)
    if kwargs.get('_recursion', False):
      raise RecursionError
    try:
      castValue = fieldType(value)
    except Exception as exception:
      raise TypeException('value', value, fieldType, ) from exception
    else:
      return self.__instance_set__(castValue, _recursion=True)

  def __instance_delete__(self, **kwargs) -> None:
    """
    AttriBox sets the value to the 'DELETED' sentinel value.
    """
    setattr(self.instance, self.getPrivateName(), DELETED)



"""HashMismatch is raised by the dispatcher system to indicate a hash
based mismatch between a type signature and a tuple of arguments. """
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from . import _Attribute
from ..text import monoSpace

try:
  from typing import TYPE_CHECKING
except ImportError:  # pragma: no cover
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from worktoy.static import TypeSig

  from typing import Self


class HashMismatch(Exception):
  """HashMismatch is raised by the dispatcher system to indicate a hash
  based mismatch between a type signature and a tuple of arguments. """

  typeSig = _Attribute()
  posArgs = _Attribute()
  types = _Attribute()

  def __init__(self, typeSig_: TypeSig, *args) -> None:
    """HashMismatch is raised by the dispatcher system to indicate a hash
    based mismatch between a type signature and a tuple of arguments. """
    self.types = typeSig_
    self.posArgs = args

    Exception.__init__(self, )

  def __str__(self) -> str:
    """Get the string representation of the HashMismatch."""
    sigStr = str(self.typeSig)
    argTypes = [type(arg).__name__ for arg in self.posArgs]
    argStr = """(%s)""" % ', '.join(argTypes)
    sigHash = hash(self.typeSig)
    try:
      argHash = hash(self.posArgs)
    except TypeError:
      argHash = '<unhashable>'

    infoSpec = """Unable to match type signature: <br><tab>%s<br>with
    signature of arguments:<br><tab>%s<br>Received hashes: %d != %s"""
    info = infoSpec % (sigStr, argStr, sigHash, argHash)
    return monoSpace(info)

  __repr__ = __str__



"""
This file provides the two private descriptors representing the top-level
instance and owner of the descriptor owning the hook that owns this
descriptor.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ..static import AbstractObject

try:
  from typing import TYPE_CHECKING
except ImportError:  # pragma: no cover
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any, Callable, Self


class _TopInstance(AbstractObject):
  """
  This private descriptor reflects the top-level instance of the descriptor
  owning the hook that owns this descriptor.
  """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  GETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __instance_get__(self, **kwargs) -> Any:
    """
    Returns the top-level instance of the descriptor owning the hook.
    """
    return self.instance.instance


class _TopOwner(AbstractObject):
  """
  This private descriptor reflects the top-level owner of the descriptor
  owning the hook that owns this descriptor.
  """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  GETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __instance_get__(self, **kwargs) -> Any:
    """
    Returns the top-level instance of the descriptor owning the hook.
    """
    return self.instance.owner



"""DataField represents an entry in the EZData classes. """
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from worktoy.waitaminute import TypeException

try:
  from typing import TYPE_CHECKING
except ImportError:  # pragma: no cover
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False


class DataField:
  """DataField represents an entry in the EZData classes. """
  __slots__ = ('key', 'type_', 'val')

  def __init__(self, key: str, type_: type, val: object) -> None:
    """Initialize the DataField object."""
    if not isinstance(key, str):
      raise TypeException('key', key, str)
    if not isinstance(type_, type):
      raise TypeException('type_', type_, type)
    if not isinstance(val, type_):
      raise TypeException('val', val, type_)
    self.key = key
    self.type_ = type_
    self.val = val

  def __str__(self) -> str:
    """Get the string representation of the DataField object."""
    infoSpec = """%s: %s = %s(%s)"""
    clsName = type(self).__name__
    typeName = self.type_.__name__
    return infoSpec % (self.key, typeName, clsName, self.val)

  def __repr__(self) -> str:
    """
    Get the code that would create this DataField object.
    """
    infoSpec = """%s(%r, %r, %r)"""
    return infoSpec % (type(self).__name__, self.key, self.type_, self.val)



"""
NestBox subclasses 'AttriBox' and provides a descriptor intended for use
by other descriptors.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from . import AttriBox, _TopInstance, _TopOwner
from ..static import Alias, AbstractObject
from ..static.zeroton import DELETED
from ..waitaminute import TypeException

try:
  from typing import TYPE_CHECKING
except ImportError:  # pragma: no cover
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any, Callable, Self


class NestBox(AttriBox):
  """
  NestBox subclasses 'AttriBox' and provides a descriptor intended for use
  by other descriptors. This means that the object at the 'instance'
  attribute is itself a descriptor and the actual instance is that of the
  owning instance. To facilitate this, the 'Alias' descriptor adds 'desc'
  and 'descType' as alternative names for the 'instance' and 'owner'
  respectively. Finally, the 'topInstance' and 'topOwner' attributes
  are provided to access the top-level instance and owner of the descriptor.
  Those are the objects that are not descriptors.

  - instance: The immediate owning instance of this descriptor.
  - owner: The immediate owning class of this descriptor.
  - desc: An alias for 'instance'
  - descType: An alias for 'owner'
  - topInstance: The top-level instance of the descriptor.
  - topOwner: The top-level owner of the descriptor.
  """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Public Variables
  desc = Alias('instance')
  descType = Alias('owner')
  topInstance = _TopInstance()
  topOwner = _TopOwner()

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  DOMAIN SPECIFIC  # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def _createFieldObject(self, ) -> Any:
    """
    Creates the field object using the type specified in the AttriBox.
    This method is used by the 'strongBox' to instantiate the field
    object when it is not already cached.
    """
    fieldType = self.getFieldType()
    ins, own, desc = self.topInstance, self.topOwner, self.desc
    args = self._getPositionalArgs(THIS=ins, OWNER=own, DESC=desc)
    kwargs = self._getKeywordArgs(THIS=ins, OWNER=own, DESC=desc)
    return fieldType(*args, **kwargs)

  def __instance_get__(self, **kwargs, ) -> Any:
    """
    Attempts to retrieve the value of the field object from the instance.
    If the value is 'None', or if 'getattr' raises 'AttributeError',
    a new 'fieldObject' is created and set on the instance and the
    method recurses to retrieve the value again. This ensures that the
    retrieval does work.
    """
    pvtName = self.getPrivateName()
    try:
      value = getattr(self.topInstance, pvtName, )
    except AttributeError:
      if kwargs.get('_recursion', False):
        raise RecursionError
      value = self._createFieldObject()
      setattr(self.topInstance, pvtName, value)
      return self.__instance_get__(_recursion=True)
    else:
      if value is None:
        if kwargs.get('_recursion2', False):
          raise RecursionError
        value = self._createFieldObject()
        setattr(self.topInstance, pvtName, value)
        return self.__instance_get__(_recursion2=True)
      return value

  def __instance_set__(self, value: Any, **kwargs) -> None:
    """
    Implements type guard.
    """
    fieldType = self.getFieldType()
    if isinstance(value, fieldType):
      pvtName = self.getPrivateName()
      return setattr(self.topInstance, pvtName, value)
    if kwargs.get('_recursion', False):
      raise RecursionError
    try:
      castValue = fieldType(value)
    except Exception as exception:
      raise TypeException('value', value, fieldType, ) from exception
    else:
      return self.__instance_set__(castValue, _recursion=True)

  def __instance_delete__(self, **kwargs) -> None:
    """
    AttriBox sets the value to the 'DELETED' sentinel value.
    """
    setattr(self.topInstance, self.getPrivateName(), DELETED)



"""THIS is the token object indicating a class before it is created. """
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ...text import monoSpace

from . import Zeroton

try:
  from typing import TYPE_CHECKING
except ImportError:  # pragma: no cover
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  pass


class THIS(metaclass=Zeroton):
  """THIS is the token object indicating a class before it is created. """

  @classmethod
  def __class_str__(cls, ) -> str:
    """Return the class name of the class."""
    infoSpec = """The '%s' Zeroton indicates a placeholder for a class not 
    yet created allowing it to be referenced before creation. """
    return monoSpace(infoSpec % cls.__name__)



"""_Attribute provides a primitive descriptor protocol implementation. """
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

import re

try:
  from typing import TYPE_CHECKING
except ImportError:  # pragma: no cover
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  pass


class _Meta(type):
  """Metaclass simplifying the __str__ of classes. """

  def __str__(cls, ) -> str:
    """
    Return a string representation of the class.
    """
    clsName = cls.__name__
    metaName = type(cls).__name__
    infoSpec = """%s(metaclass=%s)""" % (clsName, metaName,)
    return infoSpec

  def __repr__(cls, ) -> str:
    """
    Return a string representation of the class.
    """
    clsName = cls.__name__
    metaName = type(cls).__name__
    baseNames = [base.__name__ for base in cls.__bases__]
    baseStr = ', '.join(baseNames)
    nameSpace = ['%s=%s' % (k, v) for k, v in cls.__dict__.items()]
    nameStr = ', '.join(nameSpace)
    infoSpec = """%s('%s', (%s), dict(%s)"""
    return infoSpec % (clsName, metaName, baseStr, nameStr,)


class _Attribute(metaclass=_Meta):
  """_Attribute provides a primitive descriptor protocol implementation.
  This is a base class for all attributes. It provides the basic
  functionality
  for getting and setting values, as well as the ability to define custom
  behavior for getting and setting values.
  """

  __field_name__ = None
  __field_owner__ = None

  __private_name__ = None
  __default_value__ = None

  def __set_name__(self, owner: type, name: str) -> None:
    """Set the name of the attribute."""
    self.__field_name__ = name
    self.__field_owner__ = owner

  def _getPrivateName(self) -> str:
    """Get the private name of the attribute."""
    pattern = re.compile(r'(?<!^)(?=[A-Z])')
    return '__%s__' % pattern.sub('_', self.__field_name__).lower()

  def __init__(self, *args) -> None:
    """Initialize the attribute with a default value."""
    if args:
      if len(args) == 1:
        self.__default_value__ = args[0]
      else:
        self.__default_value__ = (*args,)

  def __get__(self, instance: object, owner: type) -> object:
    """Get the value of the attribute."""
    if instance is None:
      return self
    privateName = self._getPrivateName()
    return getattr(instance, privateName, self.__default_value__)

  def __set__(self, instance: object, value: object) -> None:
    """Set the value of the attribute."""
    privateName = self._getPrivateName()
    setattr(instance, privateName, value)

  def __delete__(self, instance: object) -> None:
    """Delete the value of the attribute."""
    privateName = self._getPrivateName()
    delattr(instance, privateName)

  def __str__(self, ) -> str:
    """
    Return a string representation of the attribute.
    """
    ownerName = self.__field_owner__.__name__
    fieldName = self.__field_name__
    clsName = type(self).__name__
    infoSpec = """%s(%s, %s)""" % (clsName, ownerName, fieldName,)
    return infoSpec

  def __repr__(self, ) -> str:
    """
    Return a string representation of the attribute.
    """
    ownerName = self.__field_owner__.__name__
    fieldName = self.__field_name__
    clsName = type(self).__name__
    infoSpec = """%s = %s(%s)"""
    if self.__default_value__ is None:
      defaultStr = ''
    defaultStr = repr(self.__default_value__)
    return infoSpec % (fieldName, clsName, defaultStr,)



"""The 'worktoy.mcls' module provides the base custom metaclass and
namespace uses across the 'worktoy' library. The pattern used is for the
namespace class to implement a method called 'compile' which decides the
actual namespace to be used in the class creation based on the code found
during execution of the class body.

The metaclass instantiates the namespace class and returns the created
namespace object from the '__prepare__' method. Upon completion of the
class body execution this namespace object is passed back to the metaclass
in the '__new__' method. The namespace object is required to implement a
method called 'compile' which returns an instance of 'dict'. Next the
metaclass validates the 'dict' object and finally passes it to the
'__new__' method on type. Before returning the created class,
the metaclass checks each baseclass for the presence of a method called
'__subclasshook__'. If it exists, the method is called with the created
class object allowing the baseclass to modify or even reject the class.
Finally, the metaclass returns the created class object."""
#  AGPL-3.0 license
#  Copyright (c) 2024-2025 Asger Jon Vistisen
from __future__ import annotations

from . import hooks

from ._nuthin import newBuild as _newBuild  # Private
from ._type_names import Spaces, Base, Space, Types
from ._abstract_namespace import AbstractNamespace
from ._abstract_metaclass import AbstractMetaclass
from ._base_space import BaseSpace
from ._base_meta import BaseMeta
from ._base_object import BaseObject

__all__ = [
    'hooks',
    'Spaces',
    'Base',
    'Space',
    'Types',
    'AbstractNamespace',
    'AbstractMetaclass',
    'BaseSpace',
    'BaseMeta',
    'BaseObject',
]



"""FloatCast converts from str, int, bool and complex to float."""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ...waitaminute import CastMismatch
from . import AbstractCast

try:
  from typing import TYPE_CHECKING
except ImportError:  # pragma: no cover
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Callable


class FloatCast(AbstractCast):
  """FloatCast converts from str, int, bool and complex to float."""

  def getTargetType(self, **kwargs) -> type:
    """Get the target type of the cast. Subclasses must implement this
    method to return the target type."""
    return float

  def _cast(self, *args, **kwargs) -> float:
    """Cast the arguments to the target type."""
    if len(args) - 1:
      e = """%s expects exactly one argument, but received %d: \n%s"""
      clsName = type(self).__name__
      nArgs = len(args)
      fmt = lambda arg: '  %s of type: %s' % (str(arg), type(arg).__name__)
      listArgs = '\n'.join([fmt(arg) for arg in args])
      raise TypeError(e % (clsName, nArgs, listArgs))
    arg = args[0]
    if isinstance(arg, float):
      return arg
    if isinstance(arg, bool):
      return 1.0 if arg else 0.0
    if isinstance(arg, int):
      return float(arg)
    if isinstance(arg, complex):
      if arg.imag:
        raise CastMismatch(float, arg)
      return float(arg.real)
    try:
      return float(arg)
    except Exception as exception:
      raise CastMismatch(float, arg) from exception

  def __get__(self, instance: object, owner: type) -> Callable:
    """Get the FloatCast object."""
    return self



"""ReservedNames provides a list of reserved names that are set
automatically by the interpreter. """
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ...text import monoSpace
from ...waitaminute import ReadOnlyError

try:
  from typing import TYPE_CHECKING
except ImportError:  # pragma: no cover
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Self, Any, Iterator


class _Meta(type):
  """Metaclass for ReservedNames."""

  __reserved_names__ = [
      '__dict__',
      '__weakref__',
      '__module__',
      '__annotations__',
      '__match_args__',
      '__doc__',
      '__name__',
      '__qualname__',
      '__firstlineno__',
      '__static_attributes__',
  ]

  def __iter__(cls) -> Iterator[str]:
    """Iterate over the reserved names."""
    yield from cls.__reserved_names__

  def __call__(cls, *args, **kwargs) -> Self:
    """Call the metaclass."""
    return cls

  def __get__(cls, instance: object, owner: type) -> Self:
    """Get the reserved names."""
    return cls

  def __set__(cls, instance: object, value: Any) -> None:
    """Set the reserved names."""
    raise ReadOnlyError(instance, cls, value)

  def __delete__(cls, instance: object) -> None:
    """Delete the reserved names."""
    raise ReadOnlyError(instance, cls, None)

  def __contains__(cls, name: str) -> bool:
    """Check if the name is in the reserved names."""
    return True if name in [i for i in cls] else False

  def __str__(cls, ) -> str:
    """Get the string representation of the metaclass."""
    info = 'ReservedNames:\n%s'
    names = '<br><tab>'.join([name for name in cls])
    return monoSpace(info % names)

  __repr__ = __str__


class ReservedNames(metaclass=_Meta):
  """ReservedNames provides a list of reserved names that are set
  automatically by the interpreter."""
  pass



"""
ZeroSpace provides the namespace objects for the Zeroton metaclass.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ...waitaminute import ReservedName
from .. import HistDict
from . import _reservedNames

try:
  from typing import TYPE_CHECKING
except ImportError:  # pragma: no cover
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import TypeAlias, Any

  Bases: TypeAlias = tuple[type, ...]
  Space: TypeAlias = dict[str, Any]


class ZeroSpace(HistDict):
  """
  ZeroSpace provides the namespace objects for the Zeroton metaclass.
  """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  CONSTRUCTORS   # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  @staticmethod
  def _isRoot(obj: object) -> bool:
    """
    Check if the object is a root object.
    """
    trust = getattr(obj, '__trust_me_bro__', None)
    return False if trust is None else (True if trust else False)

  def __init__(self, mcls: type, name: str, bases: Bases, **kwargs) -> None:
    """
    Initialize the ZeroSpace with the given arguments.
    """
    self.__meta_class__ = mcls
    self.__class_name__ = name
    self.__base_classes__ = bases
    self.__key_args__ = kwargs.get('keyArgs', None)

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  Python API   # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __setitem__(self, key: str, value: object, **kwargs) -> None:
    """
    Set the item in the ZeroSpace.
    """
    if key in _reservedNames and not self._isRoot(value):
      raise ReservedName(key)
    return HistDict.__setitem__(self, key, value)

  def compile(self, ) -> Space:
    """
    Compiles the namespace to an instance of a traditional 'dict'.
    """
    out = dict()
    for call in self.getItemCalls():
      call.apply(out)
    return out



"""
The 'auto' function conveniently specifies a member of a KeeNum
enumeration right in the class body.

When subclasses of KeeNum are created, the 'auto' function collects
positional and keyword arguments later used to instantiate the member
type to create the enumeration member value.


"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ..static import AbstractObject

try:
  from typing import TYPE_CHECKING
except ImportError:  # pragma: no cover
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any


class _AutoMember(AbstractObject):
  """
  _AutoMember encapsulates future enumeration member objects. The 'auto'
  function instantiates this class during the class body execution. The
  control flow collects these objects during compilation of the final
  namespace object. For this reason, _AutoMember is private to the
  'worktoy.keenum' module. Instances of _AutoMember are discarded during the
  class creation process.
  """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Private variables
  __auto_member__ = True

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  GETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  getKeywordArgs = AbstractObject._getKeywordArgs  # Going public
  getPositionalArgs = AbstractObject._getPositionalArgs  # Going public

  def getValue(self) -> Any:
    """
    Returns the first positional argument
    """
    return (self.getPositionalArgs() or [None, ])[0]


def auto(*args: Any, **kwargs: Any) -> AbstractObject:
  """
  The 'auto' function conveniently specifies a member of a KeeNum
  enumeration right in the class body.

  When subclasses of KeeNum are created, the 'auto' function collects
  positional and keyword arguments later used to instantiate the member
  type to create the enumeration member value.
  """

  return _AutoMember(*args, **kwargs, )



"""The 'validateExistingDirectory' function validates that a given 'str'
object points to an existing directory. """
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

import os

from worktoy.text import monoSpace

try:
  from typing import TYPE_CHECKING
except ImportError:  # pragma: no cover
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any, Optional, Union, Self, Callable, TypeAlias


def validateExistingDirectory(directory: str) -> str:
  """
  Validates that a given 'str' object points to an existing directory.

  Args:
    directory (str): The directory to validate.

  Returns:
    str: The validated directory.

  Raises:
    FileNotFoundError: If the directory does not exist.
    NotADirectoryError: If the path is not a directory.
  """
  if not os.path.exists(directory):
    infoSpec = """No directory exists at: '%s'!"""
    info = monoSpace(infoSpec % directory)
    raise FileNotFoundError(info)
  if not os.path.isdir(directory):
    infoSpec = """The path '%s' is not a directory!"""
    info = monoSpace(infoSpec % directory)
    raise NotADirectoryError(info)
  return os.path.normpath(directory)



"""The 'worktoy.work_io' module provides I/O functionalities. """
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ._validate_existing_directory import validateExistingDirectory
from ._validate_existing_file import validateExistingFile
from ._validate_available_path import validateAvailablePath
from ._fid_gen import FidGen
from ._new_directory import newDirectory
from ._yeet_directory import yeetDirectory

__all__ = [
    'validateExistingDirectory',
    'validateExistingFile',
    'validateAvailablePath',
    'FidGen',
    'newDirectory',
    'yeetDirectory',
]



"""
The 'worktoy.core' module provides the most primitive objects used by the
'worktoy' library.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ._some import Some
from ._unpack import unpack
from ._bipartite_matching import bipartiteMatching

__all__ = [
    'Some',
    'unpack',
    'bipartiteMatching',
]



"""
UnpackException is a custom exception raised to indicate that an unpacking
operation found no argument requiring unpacking. This is raised by the
'unpack' function in the 'worktoy.core' module when the 'strict' mode is
enabled and no iterable is found among the arguments. Please note that
this function does not consider strings or bytes as unpackable iterables.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from . import _Attribute

try:
  from typing import TYPE_CHECKING
except ImportError:  # pragma: no cover
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Self, TypeAlias, Any


class UnpackException(ValueError):
  """
  UnpackException is a custom exception raised to indicate that an unpacking
  operation found no argument requiring unpacking. This is raised by the
  'unpack' function in the 'worktoy.core' module when the 'strict' mode is
  enabled and no iterable is found among the arguments. Please note that
  this function does not consider strings or bytes as unpackable iterables.
  """

  posArgs = _Attribute()  # type: tuple[Any, ...]

  def __init__(self, *args: Any) -> None:
    """
    Initialize the UnpackException with an optional message.
    """
    self.posArgs = args
    ValueError.__init__(self, )

  def __str__(self) -> str:
    """
    Return a string representation of the UnpackException.
    """
    infoSpec = """'unpack' found no iterable argument from: \n'%s'\nand is 
    running in strict mode (default). Change this by setting keyword 
    argument 'strict' to False."""
    argStr = '\n  '.join([str(arg) for arg in self.posArgs])
    info = infoSpec % argStr
    return info

  __repr__ = __str__



"""
IllegalDispatcher is a custom exception raised to indicate that a subclass
of 'Dispatch' has failed to identify itself as a dispatcher. All
subclasses of 'Dispatch' must provide the '__overload_dispatcher__' key in
their namespace allowing 'TypeSig' to identify recursive constructor calls
that would otherwise cause hard-to-debug recursion errors. These errors
might even include infinite recursions not detectable by the Python
recursion limit.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from . import _Attribute
from ..text import monoSpace

try:
  from typing import TYPE_CHECKING
except ImportError:  # pragma: no cover
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  pass


class IllegalDispatcher(Exception):
  """IllegalDispatcher is a custom exception raised to indicate that a
  subclass of 'Dispatch' has failed to identify itself as a dispatcher. All
  subclasses of 'Dispatch' must provide the '__overload_dispatcher__' key in
  their namespace allowing 'TypeSig' to identify recursive constructor calls
  that would otherwise cause hard-to-debug recursion errors. These errors
  might even include infinite recursions not detectable by the Python
  recursion limit.
  """

  dispatcher = _Attribute()

  def __init__(self, cls: type) -> None:
    """Initialize the IllegalDispatcher exception."""
    Exception.__init__(self, )
    self.dispatcher = cls

  def __str__(self, ) -> str:
    """
    Return a string representation of the IllegalDispatcher exception.
    """
    infoSpec = """IllegalDispatcher: Class '%s' failed to set 
      '__overload_dispatcher__ = True'. This token is required for overload 
      dispatchers so that TypeSig can prevent infinite recursion during 
      __init__ overload resolution."""
    return monoSpace(infoSpec % self.dispatcher.__name__)

  __repr__ = __str__



"""
The 'Some' class considers all objects as instances of itself, except for
'None'.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

try:
  from typing import TYPE_CHECKING
except ImportError:  # pragma: no cover
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import TypeAlias, Any

  Bases: TypeAlias = tuple[type, ...]
  Space: TypeAlias = dict[str, Any]


class _InitSub(object):
  """
  A chill object that does not raise any:
  'TypeError: Some.__init_subclass__() takes no keyword arguments'
  """

  def __init__(self, *args, **kwargs) -> None:
    """
    Why are we still here?
    """
    object.__init__(self)

  def __init_subclass__(cls, **kwargs) -> None:
    """
    Just to suffer?
    """
    object.__init_subclass__()


class _MetaSome(type):
  """
  Implementation of custom 'isinstance' logic requires metaclass level
  implementation. This metaclass implements the logic that considers all
  objects as instances of 'Some', except for 'None'. It also prevents
  'Some' from being subclassed and from being instantiated.
  """

  def __instancecheck__(*args) -> bool:  # NOQA
    """
    Custom 'isinstance' logic that considers all objects as instances of
    'Some', except for 'None'.
    """
    for arg in args:
      if arg is None:
        return False
    else:
      return True

  def __subclasscheck__(cls, other: type) -> bool:  # NOQA
    """
    Any class except 'NoneType' is considered a subclass of 'Some'.
    """
    try:
      _ = issubclass(other, object)  # even NoneType is an object
    except Exception as exception:
      raise exception
    else:
      if other is type(None):
        return False
      return True

  @classmethod
  def __prepare__(mcls, name: str, bases: Bases, **kwargs) -> Space:
    """
    Prepares the namespace for the 'Some' class, ensuring it is not
    instantiated or subclassed.
    """
    if kwargs.get('_root', False):
      return dict()
    from worktoy.waitaminute import IllegalInstantiation
    raise IllegalInstantiation(mcls)

  def __new__(mcls, name: str, bases: tuple, *args, **kwargs) -> type:
    """
    Inserts the chill '_InitSub' class into the bases to prevent the cringy:
    'TypeError: Some.__init_subclass__() takes no keyword arguments'
    """
    return type.__new__(mcls, name, (*bases, _InitSub), *args, **kwargs)


class Some(metaclass=_MetaSome, _root=True):
  """
  The 'Some' class considers all objects as instances of itself, except for
  'None'. It is not meant to be instantiated or subclassed.
  """



"""The 'worktoy.workWrap' function receives an integer defining character
width and any number of strings. The function then returns a list of
strings containing the words from the strings received such that each
entry in the list does not exceed the character width. """
#  AGPL-3.0 license
#  Copyright (c) 2024-2025 Asger Jon Vistisen
from __future__ import annotations

from worktoy.text import typeMsg


def _wrap(width: int, *textLines) -> list[str]:
  """The wordwrap function wraps the input text to a specified width."""
  words = []
  for line in textLines:
    words.extend(line.split())
  lines = []
  line = []
  while words:
    word = words.pop(0)
    if len(' '.join([*line, word])) <= width:
      line.append(word)
    else:
      lines.append(' '.join(line))
      line = [word]
  if line:
    lines.append(' '.join(line))
  return lines


def _splitWords(*textLines) -> list[str]:
  """The wordwrap function wraps the input text to a specified width."""
  words = []
  for line in textLines:
    words.extend(line.split())
  return words


def _combineWords(width: int, *words) -> list[str]:
  """The wordwrap function wraps the input text to a specified width."""
  lines = []
  line = []
  words = [*words, ]
  while words:
    word = words.pop(0)
    if len(' '.join([*line, word])) <= width:
      line.append(word)
    else:
      lines.append(' '.join(line))
      line = [word]
  return lines


def wordWrap(width: int, *textLines, **kwargs) -> str:
  """The wordwrap function wraps the input text to a specified width."""
  newLine = kwargs.get('newLine', '<br>').strip().lower()
  if not isinstance(width, int):
    raise TypeError(typeMsg('width', width, int))
  words = []
  for line in textLines:
    if not isinstance(line, str):
      raise TypeError(typeMsg('line', line, str))
    words.extend(line.split())
  lines = []
  line = []
  while words:
    word = words.pop(0)
    if not word:
      continue
    if word.lower() == newLine or len(' '.join([*line, word])) > width:
      lines.append(' '.join(line))
      line = []
      continue
    line.append(word)
  return '\n'.join(lines)



"""ZerotonCaseException is a custom exception class raised when an attempt
is made to create a Zeroton class with a name not in all upper cases. """
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from . import _Attribute

try:
  from typing import TYPE_CHECKING
except ImportError:  # pragma: no cover
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Never


class ZerotonCaseException(ValueError):
  """ZerotonCaseException is a custom exception class raised when an attempt
  is made to create a Zeroton class with a name not in all upper cases. """

  name = _Attribute()

  def __init__(self, name: str) -> None:
    """Initialize the ZerotonCaseException with the name."""
    self.name = name
    ValueError.__init__(self, )

  def __str__(self) -> str:
    """String representation of the exception."""
    if self.name is None:
      return ValueError.__str__(self)
    infoSpec = """Zeroton class name '%s' must be in all upper cases!"""
    return infoSpec % self.name

  __repr__ = __str__



"""
The 'worktoy.mcls.hooks' package provides the hooks used by the
Namespace system.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ._reserved_names import ReservedNames
from ._abstract_hook import AbstractHook
from ._reserved_name_hook import ReservedNameHook
from ._pre_class_hook import PreClassHook
from ._overload_hook import OverloadHook
from ._name_hook import NameHook

__all__ = [
    'ReservedNames',
    'AbstractHook',
    'ReservedNameHook',
    'PreClassHook',
    'OverloadHook',
    'NameHook',
]



"""The 'worktoy.text' package provides functions related to text.

- 'monoSpace': Replaces all continuous whitespace with a single space.
- 'typeMsg': Creates a common type error message
- 'joinWords': Joins words with a specified separator.
- 'stringList': Creates a list of strings from a given string.
- 'wordWrap': Wraps text to a specified width.
- 'funcReport': Generates a report of the function's arguments and their
  types.



"""
#  AGPL-3.0 license
#  Copyright (c) 2024-2025 Asger Jon Vistisen
from __future__ import annotations

from ._class_body_template import ClassBodyTemplate
from ._mono_space import monoSpace
from ._join_words import joinWords
from ._type_msg import typeMsg
from ._string_list import stringList
from ._word_wrap import wordWrap
from ._func_report import funcReport

__all__ = [
    'ClassBodyTemplate',
    'monoSpace',
    'typeMsg',
    'joinWords',
    'stringList',
    'wordWrap',
    'funcReport',
]



"""FidGen provides filename generator. Given a format specification and a
directory, it returns the next available filename of the given format."""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

import os

from worktoy.attr import Field
from worktoy.mcls import BaseObject
from worktoy.parse import maybe
from worktoy.static import overload
from worktoy.text import stringList
from worktoy.waitaminute import TypeException, MissingVariable
from worktoy.work_io import validateExistingDirectory

try:
  from typing import TYPE_CHECKING
except ImportError:  # pragma: no cover
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any, Optional


class FidGen(BaseObject):
  """
  FidGen provides filename generator. Given a format specification and a
  directory, it returns the next available filename of the given format.
  """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Fallback variables
  __fallback_extension__ = 'json'
  __fallback_name__ = 'untitled'
  __fallback_directory__ = Field()

  #  Private variables
  __base_name__ = None
  __file_extension__ = None
  __file_directory__ = None

  #  Public variables
  fileExtension = Field()
  fileDirectory = Field()
  baseName = Field()

  #  Virtual variables
  filePath = Field()

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  GETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  @__fallback_directory__.GET
  def _getFallbackDirectory(self, **kwargs) -> str:
    """
    Getter-function for the fallback directory.
    """
    return os.getcwd()

  @baseName.GET
  def _getBaseName(self, ) -> str:
    """Get the name of the file."""
    if isinstance(self.__base_name__, str):
      return self.__base_name__
    if isinstance(self.owner, type):
      return self.owner.__name__
    return self.__fallback_name__

  @fileExtension.GET
  def _getFileExtension(self, **kwargs) -> str:
    """Get the file extension."""
    return maybe(self.__file_extension__, self.__fallback_extension__, )

  @fileDirectory.GET
  def _getFileDirectory(self, **kwargs) -> str:
    """Get the file directory."""
    return maybe(self.__file_directory__, self.__fallback_directory__, )

  @filePath.GET
  def _getFilePath(self, **kwargs) -> str:
    """Getter-function for the file path. """
    n = kwargs.get('_n', 0)
    formatSpec = """%s_%03d"""
    fid = formatSpec % (self.baseName, n,)
    for item in os.listdir(self.fileDirectory):
      if item.startswith(fid):
        return self._getFilePath(_n=n + 1, )
    fidExt = """%s.%s""" % (fid, self.fileExtension)
    out = os.path.join(self.fileDirectory, fidExt)
    return str(out)

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  SETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  Setter methods
  @baseName.SET
  def _setBaseName(self, newName: str) -> None:
    """Set the name of the file."""
    if self.__base_name__ == newName:
      return
    if not isinstance(newName, str):
      raise TypeException('__name__', newName, str)
    if not newName:
      raise ValueError('__name__ must be a non-empty string')
    self.__base_name__ = newName

  @fileExtension.SET
  def _setFileExtension(self, value: str) -> None:
    """Set the file extension."""
    if not isinstance(value, str):
      raise TypeException('__file_extension__', value, str)
    if not value:
      raise ValueError('__file_extension__ must be a non-empty string')
    self.__file_extension__ = value

  @fileDirectory.SET
  def _setFileDirectory(self, value: str, **kwargs) -> None:
    """Set the file directory."""
    if not isinstance(value, str):
      raise TypeException('__file_directory__', value, str)
    if not value:
      raise ValueError('__file_directory__ must be a non-empty string')
    try:
      validateExistingDirectory(value)
    except FileNotFoundError as e:
      if kwargs.get('_recursion', False):
        raise RecursionError from e
      os.makedirs(value, exist_ok=True)
      return self._setFileDirectory(value, _recursion=True)
    else:
      self.__file_directory__ = value
    finally:
      pass  # NOQA

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  CONSTRUCTORS   # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  @overload(str)
  def __init__(self, fileName: str, **kwargs) -> None:
    """Initialize the FidGen object."""
    self.__base_name__ = fileName
    self.__init__(**kwargs)

  @overload(str, str, str)
  @overload(str, str)
  def __init__(self, *args, **kwargs) -> None:
    argDir = self._findDirectory(*args)
    argExt = self._findFileExtension(*args)
    for arg in args:
      if arg in [argDir, argExt]:
        continue
      self.baseName = arg
      break
    if argDir is not None:
      self.fileDirectory = argDir
    if argExt is not None:
      self.fileExtension = argExt
    self.__init__(**kwargs)

  @overload()  # kwargs
  def __init__(self, **kwargs) -> None:
    """Initialize the FidGen object."""
    nameKeys = stringList("""name, file, fileName, filename, file_name""")
    extKeys = stringList("""ext, extension, file_extension""")
    dirKeys = stringList("""dir, directory, file_directory""")
    KEYS = [nameKeys, extKeys, dirKeys]
    NAMES = stringList("""name, ext, dirPath""")
    TYPES = dict(name=str, ext=str, dirPath=str)
    VALUES = dict()
    for (keys, (name, type_)) in zip(KEYS, TYPES.items()):
      for key in keys:
        if key in kwargs:
          value = kwargs[key]
          if isinstance(value, type_):
            VALUES[name] = value
            break
          raise TypeException(key, value, type_)

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  Python API   # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  DOMAIN SPECIFIC  # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  @staticmethod
  def _findDirectory(*args) -> Optional[str]:
    """Finds the directory in positional arguments. """
    for arg in args:
      if isinstance(arg, str):
        if os.path.isabs(arg):
          return arg
    return None  # pycharm, please!

  @staticmethod
  def _getCommonExtensions() -> list[str]:
    """Returns a list of common file extensions."""
    return stringList(
        """json, txt, csv, xml, html, pdf, doc, csv, py, 
        mkv, mp4, mp3, wav, jpg, png, gif, zip, tar, gz, bz2"""
    )

  @classmethod
  def _findFileExtension(cls, *args) -> Optional[str]:
    """Finds the file extension in positional arguments. """
    commonExtensions = cls._getCommonExtensions()
    for arg in args:
      if isinstance(arg, str):
        if str.startswith(arg, '*.'):
          return arg[2:]
        if arg in commonExtensions:
          return arg
    return None  # pycharm, please!



"""
The 'worktoy.parse' module provides the None-aware 'maybe' function
along with some related functions. The 'maybe' function implements the
only redeeming feature of javascript:

#  Javascript
const maybe = (value, fallback) => {
  return value ?? fallback;
}
console.log(maybe(null, 69, 420)); // 42
#  Python
maybe(None, 69, 420) # 42

Besides the 'maybe' function, the module also provides the 'maybeType'
and 'maybeTypes' functions. These functions add type filtering to the
maybe function.

maybeType(targetType: type, *args: Any) -> Any
maybeTypes(targetType: type, *args: Any) -> tuple[Any, ...]
The 'maybeType' function takes a target type and any number of arguments
and returns the first of the target type. The 'maybeTypes' function
returns all arguments matching the target type in a tuple.

For example:
>>> maybe(None, 0, 69, None)  # Returns 0, the first non-None value
>>> maybeType(int, None, 0, 69, None)  # Returns 0, the first int
>>> maybeTypes(int, None, 0, 69, None)  # Returns (0, 69), all ints
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ._maybe import maybe, maybeType, maybeTypes
from ._most_specific_base import mostSpecificBase

__all__ = [
    'maybe',
    'maybeType',
    'maybeTypes',
    'mostSpecificBase',
]



"""The 'newDirectory' function creates a new directory at the specified
path. """
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

import os

from . import validateAvailablePath

try:
  from typing import TYPE_CHECKING
except ImportError:  # pragma: no cover
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Union, TypeAlias, LiteralString

  Path: TypeAlias = Union[str, bytes, LiteralString]


def newDirectory(path: Path) -> str:
  """
  Creates a new directory at the specified path.

  Args:
    path (str): The path where the new directory will be created.

  Returns:
    str: The path of the newly created directory.

  Raises:
    FileExistsError: If the directory already exists.
    NotADirectoryError: If the path is not a directory.
    PathSyntaxException: If the path is not absolute.
  """
  validateAvailablePath(path)
  os.makedirs(path, exist_ok=True)
  return os.path.normpath(path)



"""The 'yeetDirectory' function removes a directory with all contents.
Effectively the same as: 'rm -rf <directory>'."""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

import os

from ..text import monoSpace

from ..waitaminute import PathSyntaxException

from . import validateExistingDirectory

try:
  from typing import TYPE_CHECKING
except ImportError:  # pragma: no cover
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any, Optional, Union, Self, Callable, TypeAlias


def yeetDirectory(dirPath: str, **kwargs) -> None:
  """
  Removes a directory with all contents. Effectively the same as: 'rm -rf
  <directory>'.

  Args:
    dirPath (str): The path to the directory to remove.

  Returns:
    str: The path of the removed directory.

  Raises:
    FileNotFoundError: If the directory does not exist.
    NotADirectoryError: If the path is not a directory.
    PathSyntaxException: If the path is not absolute.
  """
  try:
    validateExistingDirectory(dirPath)
  except FileNotFoundError as fileNotFoundError:
    if kwargs.get('strict', True):
      raise fileNotFoundError
  except NotADirectoryError as notADirectoryError:
    infoSpec = """The path received by 'yeetDirectory': '%s' is not a 
    directory!"""
    info = monoSpace(infoSpec % dirPath)
    raise NotADirectoryError(info) from notADirectoryError
  else:
    for root, dirs, files in os.walk(dirPath, topdown=False):
      for name in files:
        os.remove(os.path.join(root, name))
      for name in dirs:
        os.rmdir(os.path.join(root, name))
    os.rmdir(dirPath)
  finally:
    if os.path.exists(dirPath):
      infoSpec = """The directory '%s' was not removed!"""
      info = monoSpace(infoSpec % dirPath)
      raise OSError(info)



"""The 'typeMsg' function generates a structured error message when an
object with a given name does not belong to the given type.

Example:

  def square(number: int) -> int:
    #  The function expects an integer as argument
    if not isinstance(number, int):
      e = typeMsg('number', number, int)
      raise TypeError(e)
    return number ** 2

  square(69.420)
"""
#  AGPL-3.0 license
#  Copyright (c) 2024-2025 Asger Jon Vistisen
from __future__ import annotations

from warnings import warn

from . import monoSpace, joinWords


def _resolveTypeNames(*types) -> str:
  """Creates the first part of the error message listing the expected type
  or types. """
  if len(types) == 1:
    if isinstance(types[0], (tuple, list)):
      return _resolveTypeNames(*types[0])
    if isinstance(types[0], type):
      expName = types[0].__name__
    elif isinstance(types[0], str):
      expName = types[0]
    else:
      raise TypeError("""Received bad arguments: %s""" % (str(types),))
    return """Expected object of type '%s'""" % (expName,)
  typeNames = []
  for type_ in types:
    if isinstance(type_, type):
      typeNames.append("""'%s'""" % type_.__name__)
    elif isinstance(type_, str):
      typeNames.append("""'%s'""" % type_)
    else:
      raise TypeError("""Received bad arguments: %s""" % (str(types),))
  infoSpec = """Expected object of any of the following types: %s"""
  typeStr = joinWords(*typeNames, sep='or')
  return monoSpace(infoSpec % (typeStr,))


def typeMsg(name: str, obj: object, *types) -> str:
  """The 'typeMsg' function generates a structured error message when an
  object with a given name does not belong to the given type."""

  prelude = _resolveTypeNames(*types)
  actName = type(obj).__name__
  infoSpec = """%s at name: '%s', but received object of type '%s' with 
  repr: '%s'"""
  info = infoSpec % (prelude, name, actName, repr(obj))
  return monoSpace(info)



"""
Alias allows a subclass to refer to a descriptor on the parent class with
a different name.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ..waitaminute import AliasException

from . import AbstractObject

try:
  from typing import TYPE_CHECKING
except ImportError:  # pragma: no cover
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any, Optional, Union, Self, Callable, TypeAlias, Never


class Alias(AbstractObject):
  """
  Alias allows a subclass to refer to a descriptor on the parent class
  with a different name.
  """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Private variables
  __original_name__ = None

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  CONSTRUCTORS   # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __init__(self, originalName: str) -> None:
    """
    Must be instantiated with the original name of the descriptor
    """
    super().__init__()
    self.__original_name__ = originalName

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  Python API   # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __set_name__(self, owner: type, name: str, **kwargs) -> None:
    """
    Set the name of the alias in the owner class.
    """
    super().__set_name__(owner, name, **kwargs)
    parentDescriptor = getattr(owner, self.__original_name__, None)
    if parentDescriptor is None:
      raise AliasException(owner, name)
    setattr(owner, name, parentDescriptor)

  def __instance_get__(self, **kwargs) -> Any:
    """
    Returns the value of the original descriptor. Please note that if
    accessed through the owning class, with instance being 'None',
    the 'Alias' object not the original object is returned.
    """
    originalDescriptor = getattr(self.owner, self.__original_name__, None)
    if originalDescriptor is None:
      raise AliasException(self.owner, self.__original_name__)
    return originalDescriptor.__get__(self.instance, self.owner, **kwargs)



"""
OverloadHook hooks into the namespace system and collects the overload
decorated methods replacing them with a dispatcher that calls the
correct method based on the arguments passed to it.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ...static import Dispatch
from . import AbstractHook

try:
  from typing import TYPE_CHECKING
except ImportError:  # pragma: no cover
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any


class OverloadHook(AbstractHook):
  """
 OverloadHook implements method overloading in the metaclass namespace
  system. It collects all functions decorated with `@overload` and
  compiles them into a single dispatcher per function name.

  ## Purpose

  During class definition, multiple functions with the same name may be
  decorated with `@overload`, each targeting a different type signature.
  These are temporarily stored with signature metadata (`__type_sigs__`)
  and marked using the `__is_overloaded__` flag.

  The OverloadHook intercepts these assignments and stores the overloads
  in a dispatch map indexed by function name and type signature.

  ## Behavior

  - **During setItemHook**:
    Detects overloaded functions and registers them with the namespace’s
    internal overload map. This stage only collects metadata.

  - **During postCompileHook**:
    Replaces each overloaded function with a `Dispatch` object that
    selects the correct overload at runtime based on argument types.
    Injects `__dispatch_names__` into the namespace for debugging or
    reflection.

  ## Usage
  To use OverloadHook, simply declare it in your namespace class:

  class Space(AbstractNamespace):  # Must inherit from AbstractNamespace
    #  Custom namespace class inheriting from AbstractNamespace
    overloadHook = OverloadHook()  # Register the hook
    preClassHook = PreClassHook()  # Register the pre-class hook

  ## Example

  The following example shows a ComplexNumber class supporting three
  constructor overloads. It uses the BaseMeta metaclass, which provides
  a namespace equipped with OverloadHook and PreClassHook for
  signature resolution and THIS substitution.

  Real and imaginary parts are declared using AttriBox, which provides
  lazy, type-enforced attribute slots. See the documentation for
  worktoy.mcls.BaseMeta and worktoy.attr.AttriBox for details.

  ___________________________________________________________________________
  ```python```
  from typing import Self

  from worktoy.mcls import BaseMeta
  from worktoy.attr import AttriBox
  from worktoy.static import overload
  from worktoy.static.zeroton import THIS

  class ComplexNumber(metaclass=BaseMeta):
    #  Custom metaclass with overload support
    #  Properties:
    REAL = AttriBox[float](0.0)
    IMAG = AttriBox[float](0.0)

    @overload(float, float)
    def __init__(self, real: float, imag: float) -> None:
      #  Initialize a complex number with real and imaginary parts.

      self.REAL = real
      self.IMAG = imag

    @overload(complex)
    def __init__(self, z: complex) -> None:
      self.__init__(z.real, z.imag)

    @overload(THIS)  # THIS is a placeholder for ComplexNumber itself
    def __init__(self, other: Self) -> None:
      self.__init__(other.REAL, other.IMAG)
  ¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨
  After class creation, ComplexNumber.__init__ becomes a dispatcher
  object rather than any of the individual function definitions above.
  When the constructor is called, it matches the argument types against
  the registered signatures and invokes the correct implementation.

  This setup supports expressive, type-driven APIs while preserving
  clean class definitions — all enabled via hook-based metaclass
  composition.
  """

  def setItemHook(self, key: str, value: Any, old: Any, ) -> bool:
    """
    Set the item hook for the namespace system. This method is called
    when an item is set in the namespace system. It collects the
    overload decorated methods and replaces them with a dispatcher that
    calls the correct method based on the arguments passed to it.
    """
    if getattr(value, '__is_overloaded__', None) is None:
      return False
    typeSigs = getattr(value, '__type_sigs__', None)
    for sig in typeSigs:
      self.space.addOverload(key, sig, value)
    return True

  def postCompileHook(self, compiledSpace: dict) -> dict:
    """
    Post compile hook for the namespace system. This method is called
    after the namespace system is compiled. It collects the overload
    decorated methods and replaces them with a dispatcher that calls the
    correct method based on the arguments passed to it.
    """
    overloadMap = self.space.getOverloadMap()  # str: dict[TypeSig, Func]
    dispatchNames = []
    for key, sigMap in overloadMap.items():
      compiledSpace[key] = Dispatch(sigMap)
      dispatchNames.append(key)
    compiledSpace['__dispatch_names__'] = dispatchNames
    return compiledSpace



"""
_Attribute provides a simple implementation of the descriptor protocol.
The primary provider of descriptors in the 'worktoy' library is the
'worktoy.attr' module which depends on this module. For this reason,
this module makes use of '_Attribute' for descriptors.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ..waitaminute import VariableNotNone

from . import AbstractObject

try:
  from typing import TYPE_CHECKING
except ImportError:  # pragma: no cover
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Self, TypeAlias, Any


class _Attribute(AbstractObject):
  """
  _Attribute instances allow a simple write-once, read-many implementation
  of the descriptor protocol. It supports a default value, which should be
  passed to the constructor. If '__get__' is called before a value has
  been set explicitly, a default value is required and will be returned.
  This writes the value permanently to the default value, meaning that the
  instance cannot later change the value of the attribute.
  """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Private Variables
  __default_value__ = None

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  DOMAIN SPECIFIC  # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __instance_get__(self, **kwargs, ) -> Any:
    """
    Uses the private name created by the 'AbstractObject' parent class.
    """
    pvtName = self.getPrivateName()
    try:
      value = getattr(self.instance, pvtName)
    except AttributeError as attributeError:
      if kwargs.get('_recursion', False):
        raise RecursionError from attributeError
      if self.__default_value__ is None:
        raise attributeError
      self.__instance_set__(self.__default_value__, )
      return self.__instance_get__(_recursion=True, )
    else:
      if value is None:
        infoSpec = """'%s' object has no attribute '%s'"""
        typeName = self.owner.__name__
        attrName = self.getFieldName()
        info = infoSpec % (typeName, attrName)
        raise AttributeError(info)
      return value

  def __instance_set__(self, value: Any, **kwargs) -> None:
    """
    Sets the value of the attribute. Please note the write-once requirement.
    """
    pvtName = self.getPrivateName()
    oldValue = getattr(self.instance, pvtName, None)
    if oldValue is not None:
      raise VariableNotNone(pvtName, oldValue)
    setattr(self.instance, pvtName, value)



"""The 'worktoy.attr' module implements the descriptor protocol."""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

#  Private descriptors for top-level instance and owner
from ._top_objects import _TopInstance, _TopOwner

#  Public descriptors for AttriBox and Field
from ._field import Field
from ._attri_box import AttriBox
from ._nest_box import NestBox

__all__ = [
    'Field',
    'AttriBox',
    'NestBox',
]



"""_CurrentOwner is a private method used by AbstractObject to expose
the current owner of the descriptor instance."""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ..waitaminute import attributeErrorFactory
from ..waitaminute import ProtectedError, ReadOnlyError

try:
  from typing import TYPE_CHECKING
except ImportError:  # pragma: no cover
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any, Never
  from . import AbstractObject


class _CurrentOwner:
  """
  _CurrentOwner is a private method used by AbstractObject to expose
  the current owner of the descriptor instance.
  """

  def __get__(self, instance: Any, owner: type) -> Any:
    """
    Return the current owner of the descriptor instance.
    """
    if instance is None:
      return self
    if TYPE_CHECKING:
      assert isinstance(instance, AbstractObject)
    if instance.__current_owner__ is None:
      if instance.__field_owner__ is None:
        return None
      return instance.__field_owner__
    return instance.__current_owner__

  def __set__(self, instance: Any, value: Any) -> Never:
    """
    This should never happen.
    """
    try:
      currentOwner = instance.__current_owner__
    except Exception as exception:
      currentOwner = exception
      raise ReadOnlyError(instance, self, None) from currentOwner
    else:
      raise ReadOnlyError(instance, self, currentOwner)

  def __delete__(self, instance: Any) -> Never:
    """
    Illegal deleter operation
    """
    try:
      currentOwner = instance.__current_owner__
    except Exception as exception:
      attributeError = attributeErrorFactory(type(self), '__current_owner__')
      raise attributeError from exception
    else:
      raise ProtectedError(instance, self, currentOwner)


class _OwnerAddress(_CurrentOwner):
  """_OwnerAddress is a private method used by AbstractObject to expose
  the address of the current owner of the descriptor instance."""

  def __get__(self, instance: Any, owner: type) -> Any:
    """
    Returns 'self' if 'instance' is 'None', otherwise returns the 'id' of
    the object at the super call.
    """
    if instance is None:
      return self
    return id(_CurrentOwner.__get__(self, instance, owner))



"""
DuplicateHookError is a custom exception raised when an attempt is made to
register a hook at a name already populated with a hook. If 'oldHook is
newHook' is True, this exception may not be appropriate.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from . import _Attribute
from ..text import monoSpace


class DuplicateHookError(Exception):
  """
  DuplicateHookError is a custom exception raised when an attempt is made to
  register a hook at a name already populated with a hook. If 'oldHook is
  newHook' is True, this exception may not be appropriate.
  """

  owner = _Attribute()
  name = _Attribute()
  existingHook = _Attribute()
  newHook = _Attribute()

  def __init__(self, *args, ) -> None:
    _owner, _name, _oldHook, _newHook = [*args, None, None, None, None][:4]
    if any(i is None for i in (_name, _oldHook, _newHook)):
      Exception.__init__(self, "Duplicate hook registration")
    else:
      self.owner = _owner
      self.name = _name
      self.existingHook = _oldHook
      self.newHook = _newHook

  def __str__(self) -> str:
    """
    String representation of the exception.
    """
    if any(i is None for i in (self.name, self.existingHook, self.newHook)):
      return Exception.__str__(self)
    infonSpec = """The class '%s' already has a hook registered 
    at name: '%s'! The existing hook is '%s', and the new hook is '%s'."""
    ownerName = self.owner.__name__
    oldHook = str(self.existingHook)
    newHook = str(self.newHook)
    info = infonSpec % (ownerName, self.name, oldHook, newHook)
    return monoSpace(info)



"""
BaseMeta provides an entry point for classes implementing the overload
protocol central to the worktoy library.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from . import AbstractMetaclass
from . import BaseSpace as BSpace
from . import Types

try:
  from typing import TYPE_CHECKING
except ImportError:  # pragma: no cover
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  pass


class BaseMeta(AbstractMetaclass):
  """
  BaseMeta is a simple extension of AbstractMetaclass that overrides
  the default class body namespace with `BaseSpace`, a custom namespace
  implementation.

  ## Purpose

  This metaclass does not introduce any behavior of its own beyond
  enabling namespace hooks defined in `BaseSpace`, such as `OverloadHook`,
  `PreClassHook`, `NameHook`, and `ReservedNameHook`.

  It serves as a ready-to-use entry point for classes that require
  hook-driven behavior during class construction without writing a custom
  metaclass.

  ## Example

      class MyClass(metaclass=BaseMeta):
        ...
  """

  @classmethod
  def __prepare__(mcls, name: str, bases: Types, **kwargs) -> BSpace:
    """Prepare the class namespace."""
    return BSpace(mcls, name, bases, **kwargs)



"""
The '_reservedNames' 'list' contains the names that classes derived from
Zeroton may not use. These names are disallowed to ensure the intended
functionality of Zeroton classes.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

_reservedNames = [
    '__new__',
    '__init__',
    '__call__',
]



"""WriteOnceError is a custom error class raised to indicate that a
variable was attempted to be written to more than once."""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ..text import monoSpace

try:
  from typing import TYPE_CHECKING
except ImportError:  # pragma: no cover
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any, Self


class WriteOnceError(Exception):
  """WriteOnceError is a custom error class raised to indicate that a
  variable was attempted to be written to more than once."""

  __var_name__ = None
  __old_value__ = None
  __new_value__ = None

  def __init__(self, name: str, oldVal: Any, newVal: Any) -> None:
    """Initialize the WriteOnceError with a name, old value and new value."""
    self.__var_name__ = name
    self.__old_value__ = oldVal
    self.__new_value__ = newVal

    infoSpec = """Attempted to overwrite variable '%s' having value '%s' to 
    new value '%s'"""
    info = infoSpec % (name, str(oldVal), str(newVal))
    Exception.__init__(self, monoSpace(info))

  def __str__(self) -> str:
    infoSpec = """Attempted to overwrite variable '%s' having value '%s' to 
    new value '%s'"""
    oldStr = str(self.__old_value__)
    newStr = str(self.__new_value__)
    info = infoSpec % (self.__var_name__, oldStr, newStr)
    return monoSpace(info)

  __repr__ = __str__



"""
_CurrentModule exposes the current module through the descriptor protocol.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

try:
  from typing import TYPE_CHECKING
except ImportError:  # pragma: no cover
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any, Never

  from . import AbstractObject


class _CurrentModule:
  """
  _CurrentModule exposes the current module through the descriptor protocol.
  """

  def __get__(self, instance: Any, owner: type, **kwargs) -> Any:
    """
    Get the current module.
    """
    if instance is None:
      return self
    return owner.__module__



"""_CurrentInstance is a private method used by AbstractObject to expose
the current owning instance of the descriptor instance."""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ..waitaminute import attributeErrorFactory
from ..waitaminute import ProtectedError, ReadOnlyError

try:
  from typing import TYPE_CHECKING
except ImportError:  # pragma: no cover
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any, Never

  from . import AbstractObject


class _CurrentInstance:
  """_CurrentInstance is a private method used by AbstractObject to
  expose
  the current owning instance of the descriptor instance."""

  def __get__(self, desc: Any, owner: type) -> Any:
    """Return the current owning instance of the descriptor instance."""
    if desc is None:
      return self
    if TYPE_CHECKING:
      assert isinstance(desc, AbstractObject)
    return desc.__current_instance__

  def __set__(self, instance: Any, value: Any) -> Never:
    """
    Raises 'ReadOnlyError'
    """
    try:
      currentInstance = instance.__current_instance__
    except Exception as exception:
      currentInstance = exception
      raise ReadOnlyError(instance, self, None) from currentInstance
    else:
      raise ReadOnlyError(instance, self, currentInstance)

  def __delete__(self, instance: Any) -> Never:
    """
    Raises 'ProtectedError'
    """
    try:
      currentInstance = instance.__current_instance__
    except Exception as exception:
      cls, name = type(instance), '__current_instance__'
      attributeError = attributeErrorFactory(cls, name)
      raise attributeError from exception
    else:
      raise ProtectedError(instance, self, currentInstance)


class _InstanceAddress(_CurrentInstance):
  """_InstanceAddress is a private method used by AbstractObject to
  expose the address of the current owning instance of the descriptor
  instance."""

  def __get__(self, desc: Any, owner: type) -> Any:
    """Return the address of the current owning instance of the
    descriptor instance."""
    if desc is None:
      return self
    return id(_CurrentInstance.__get__(self, desc, owner))



"""BadValue is a temporary custom exception used to create more
informative exceptions. """
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from . import _Attribute

try:
  from typing import TYPE_CHECKING
except ImportError:  # pragma: no cover
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any, Optional, Self


class BadValue(Exception):
  """
  BadValue is a temporary custom exception used to create more
  informative exceptions.
  """

  __wrapped__ = _Attribute()

  def __init__(self, wrapped: Exception) -> None:
    """
    Initialize the BadValue instance.
    """
    self.__wrapped__ = wrapped
    Exception.__init__(self, )

  def __eq__(self, other: Exception) -> bool:
    """
    Check if the wrapped exception is equal to another exception.
    """
    if not isinstance(other, Exception):
      return False
    cls = type(self)
    if isinstance(other, cls):
      return True if self.__wrapped__ is other.__wrapped__ else False
    return True if self.__wrapped__ is other else False

  def unwrap(self, ) -> Exception:
    """
    Unwrap the wrapped exception.
    """
    return self.__wrapped__



"""The 'funcReport' analyses a function object and returns a string
description of it including name, typehints and docstring. """
#  AGPL-3.0 license
#  Copyright (c) 2024-2025 Asger Jon Vistisen
from __future__ import annotations

from inspect import signature
from typing import get_type_hints
from types import FunctionType

from . import monoSpace

try:
  from typing import TYPE_CHECKING
except ImportError:  # pragma: no cover
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  pass


def funcReport(func: FunctionType, **kwargs) -> str:
  """The 'funcReport' analyses a function object and returns a string
  description of it including name, typehints and docstring. """

  docString = getattr(func, '__doc__', None)
  if not docString:
    if kwargs.get('strict', True):
      e = """No docstring found!"""
      raise SyntaxError(e)
    docString = """Imagine not providing a docstring!"""
  typeHints = get_type_hints(func)
  typeSig = signature(func)
  funcName = getattr(func, '__name__', 'unknown')
  #  Collecting parameter kinds from the 'inspect.Signature' object
  paramKinds = dict(oneStar='', twoStars='')
  for (key, param) in typeSig.parameters.items():
    if param.kind.name == 'VAR_POSITIONAL':
      paramKinds['oneStar'] = """*%s""" % param.name
    if param.kind.name == 'VAR_KEYWORD':
      paramKinds['twoStars'] = """**%s""" % param.name
  #  Collecting type hints
  posArgs = []
  returnType = None
  for (name, type_) in typeHints.items():
    if name == 'return':
      returnType = type_
      continue
    posArgs.append("""%s: %s""" % (name, type_.__name__))
  else:
    if returnType is None:
      if kwargs.get('strict', True):
        e = """No return type hint found!"""
        raise SyntaxError(e)
  if paramKinds['oneStar']:
    posArgs.append(paramKinds['oneStar'])
  if paramKinds['twoStars']:
    posArgs.append(paramKinds['twoStars'])
  #  Formatting the output
  argStr = ', '.join(posArgs)
  if returnType is type(None):
    outName = 'None'
  else:
    outName = returnType.__name__
  fmtSpec = """def %s(%s) -> %s:<br><tab>\"\"\"%s\"\"\""""
  out = fmtSpec % (funcName, argStr, outName, docString)
  return monoSpace(out)



"""
CascadeException provides a custom exception raised to indicate that a
system of exceptions occurred in a cascading manner after a series of
failed fallbacks. This aims at solving the same as ExceptionGroup,
but this is available only for python 3.11+.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from . import _Attribute
from ..parse import maybe
from ..text import monoSpace, wordWrap

try:
  from typing import TYPE_CHECKING
except ImportError:  # pragma: no cover
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any, Iterator, Self


class CascadeException(Exception):
  """
  CascadeException provides a custom exception raised to indicate that a
  system of exceptions occurred in a cascading manner after a series of
  failed fallbacks. This aims at solving the same as ExceptionGroup,
  but this is available only for python 3.11+.

  When instantiating provide the exceptions in the order caught.
  """

  exceptionChain = _Attribute()

  # def __new__(cls, *exceptions: Exception) -> Exception:
  #   """
  #   Create a new CascadeException instance with the given exceptions.
  #   """
  #   for exception in exceptions:
  #     if isinstance(exception, Exception):
  #       continue
  #     if isinstance(exception, BaseException):
  #       raise exception
  #   if len(exceptions) == 1:
  #     out = exceptions[0]
  #     setattr(out, '__no_init__', True)
  #     out.__cause__ = cls()
  #     return out
  #   if len(exceptions) == 2:
  #     out = exceptions[0]
  #     exceptions[1].__cause__ = cls()
  #     out.__cause__ = exceptions[1]
  #     setattr(out, '__no_init__', True)
  #     return out
  #   return Exception.__new__(cls)

  def __init__(self, *exceptions: Exception) -> None:
    """
    Initialize the CascadeException with a chain of exceptions.
    """
    self.exceptionChain = exceptions

  def __str__(self, ) -> str:
    """
    Get the string representation of the CascadeException.
    """
    if not self.exceptionChain:
      return Exception.__str__(self)
    headerSpec = """%s!\nA total of %d exceptions were raised in cascading 
    manner:"""
    clsName = type(self).__name__
    header = monoSpace(headerSpec % (clsName, len(self)))
    lines = []
    for (i, e) in enumerate(self):
      infoSpec = """%s!<br>%s"""
      head = '%s%s' % ((i + 2) * '-', type(e).__name__)
      body = wordWrap(73 - 2 * i, repr(e), ).split('\n')
      body = ('%s    <br>' % ' ' * i * 2).join([*body, ])
      lines.append(infoSpec % (head, body))
    lineStr = '<br>'.join(lines)
    return monoSpace("""%s<br>%s""" % (header, lineStr))

  def __repr__(self, ) -> str:
    """
    Code representation of the CascadeException.
    """
    body = ',<br><tab>'.join([repr(e) for e in self])
    clsName = type(self).__name__
    return monoSpace('%s(<br><tab>%s<br>)' % (clsName, body))

  def __iter__(self) -> Iterator[Exception]:
    """
    Iterate over the chain of exceptions.
    """
    for link in maybe(self.exceptionChain, []):
      yield link

  def __contains__(self, exception: Exception) -> bool:
    """
    Check if the given exception is in the chain of exceptions.
    """
    for link in self:
      if link is exception:
        return True
    else:
      return False

  def __len__(self, ) -> int:
    """
    Get the length of the chain of exceptions.
    """
    if self.exceptionChain:
      return len(self.exceptionChain)
    return 0

  def _resolveOther(self, other: Any) -> Self:
    """
    Resolve the other object to a CascadeException if possible.
    """
    cls = type(self)
    if isinstance(other, cls):
      return other
    if isinstance(other, (tuple, list)):
      for item in other:
        if not isinstance(item, Exception):
          break
      else:
        return cls(*other)
    return NotImplemented

  def __eq__(self, other: Any) -> bool:
    """
    Check if the other object is a CascadeException and has the same
    exception chain.
    """
    other = self._resolveOther(other)
    if other is NotImplemented:
      return False
    if len(self) != len(other):
      return False
    for selfLink, otherLink in zip(self, other):
      if selfLink is not otherLink:
        return False
    return True



"""ProtectedError is raised to indicate an attempt to delete a protected
object. For example, a descriptor class could implement the '__delete__'
method to always raise this exception. This provides a more detailed
error. Particularly because both TypeError and AttributeError are being
suggested by large language models. Neither of which is wrong, but lacks
the specificity of this exception.

The ProtectedError class inherits from both TypeError and AttributeError,
ensuring that it is caught in exception clauses pertaining to either.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from . import _Attribute, BadDelete
from ..text import monoSpace

try:
  from typing import TYPE_CHECKING
except ImportError:  # pragma: no cover
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any, Optional, Self


class ProtectedError(TypeError):
  """ProtectedError is raised to indicate an attempt to delete a protected
  object. For example, a descriptor class could implement the '__delete__'
  method to always raise this exception. This provides a more detailed
  error. Particularly because both TypeError and AttributeError are being
  suggested by large language models. Neither of which is wrong, but lacks
  the specificity of this exception."""

  owningInstance = _Attribute()
  descriptorObject = _Attribute()
  existingValue = _Attribute()

  def __init__(self, instance: Any, desc: Any, oldValue: Any) -> None:
    """Initialize the ReadOnlyError."""
    self.owningInstance = instance
    self.descriptorObject = desc
    self.existingValue = oldValue
    fieldOwner = getattr(instance, '__field_owner__', None)
    fieldName = getattr(desc, '__field_name__', None)
    if fieldOwner is None or fieldName is None:
      info = """Cannot delete protected attribute!"""
    else:
      fieldId = '%s.%s' % (fieldOwner.__name__, fieldName)
      if isinstance(oldValue, BadDelete) or oldValue is None:
        infoSpec = """Attempted to delete protected attribute '%s'"""
      else:
        infoSpec = """Attempted to delete protected attribute '%s' 
        with value: '%s'"""
      info = monoSpace(infoSpec % (fieldId, oldValue))
    TypeError.__init__(self, info)

  def _resolveOther(self, other: object) -> Self:
    """Resolve the other object."""
    cls = type(self)
    if isinstance(other, cls):
      return other
    if isinstance(other, (tuple, list)):
      try:
        return cls(*other)
      except TypeError:
        return NotImplemented
    return NotImplemented

  def __eq__(self, other: object) -> bool:
    """Compare the ProtectedError object with another object."""
    other = self._resolveOther(other)
    if other is NotImplemented:
      return False
    cls = type(self)
    if isinstance(other, cls):
      if self.descriptorObject != other.descriptorObject:
        return False
      if self.owningInstance != other.owningInstance:
        return False
      if self.existingValue != other.existingValue:
        return False
      return True
    return False



"""DELETED provides a value used to indicate that an attribute is to be
treated as having been deleted. For example, after a call to __delete__,
subsequent calls to __get__ should raise AttributeError until a call to
__set__ provides a new value. """
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from . import Zeroton

try:
  from typing import TYPE_CHECKING
except ImportError:  # pragma: no cover
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  pass


class DELETED(metaclass=Zeroton):
  """DELETED is a singleton class that is used to indicate that an
  attribute has been deleted. It covers an edge case where a deleted
  attribute is accessed, and this singleton is present to indicate that
  the attribute has been deleted with 'del <attr>' or similar."""
  pass



"""
The metaclass 'KeeMeta' automatically creates the KeeNum class for which
the class here is a placeholder of sorts.

"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from . import KeeMeta

KeeNum = KeeMeta.getCoreKeeNum()



"""
AbstractObject defines a base class for all other objects in the worktoy
library.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

import re

from ..text import monoSpace
from ..waitaminute import ReadOnlyError, ProtectedError, \
  attributeErrorFactory
from ..waitaminute import MissingVariable, TypeException
from ..parse import maybe
from .zeroton import THIS, OWNER, DELETED
from . import _CurrentInstance, _InstanceAddress
from . import _CurrentOwner, _OwnerAddress
from . import _CurrentClass, _CurrentModule

#  Below provides compatibility back to Python 3.7
try:
  from typing import TYPE_CHECKING
except ImportError:  # pragma: no cover
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any, TypeAlias

  Names: TypeAlias = tuple[str, str, str, str]


class AbstractObject:
  """
  AbstractObject is the base class for most core components in the
  worktoy framework. It extends Python's built-in 'object' with safer
  construction semantics, stricter descriptor behavior, and optional
  dynamic context binding.

  ## Constructor Behavior

  Both `__init__` and `__init_subclass__` are patched to discard any
  unexpected arguments. This prevents stray *args or **kwargs from
  propagating into classes that don't explicitly accept them, which is
  particularly useful in deep inheritance hierarchies.

  ## Descriptor Safety

  This class provides an extensible, context-aware descriptor pattern:

  - `__get__`, `__set__`, and `__delete__` update internal `instance` and
    `owner` references before dispatching to:
    - `__instance_get__`
    - `__instance_set__`
    - `__instance_delete__`

  Subclasses can override these to define field-like behavior, while still
  retaining contextual metadata.

  If a subclass raises `BadSet` or `BadDelete`, `__setattr__` and
  `__delattr__` catch those and re-raise them as `ReadOnlyError` or
  `ProtectedError` respectively. This provides a clean and consistent way
  to signal protected fields.

  ## Contextual Metadata

  When used as a descriptor, this class dynamically reflects runtime
  context via properties:
    - `instance`: the current instance being accessed
    - `owner`: the class owning the descriptor
    - `cls`: the descriptor's own class
    - `module`: its defining module

  These are exposed as dynamic properties for use in custom logic.

  ## Utilities

  - `parseKwargs`: searches kwargs for values of given types or names
  - `getFieldName`, `getFieldOwner`: introspective helpers
  - `getPrivateName`: converts CamelCase field names to snake_case for
    private naming

  Use AbstractObject as a foundation for anything that wants controlled
  access behavior, contextual awareness, or descriptor-like hooks.
  """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Private variables
  __field_name__ = None
  __field_owner__ = None
  __pos_args__ = None
  __key_args__ = None
  __current_owner__ = None
  __current_instance__ = None

  #  Public variables
  instance = _CurrentInstance()
  instanceId = _InstanceAddress()
  owner = _CurrentOwner()
  ownerId = _OwnerAddress()
  cls = _CurrentClass()
  module = _CurrentModule()

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  GETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def getFieldName(self) -> str:
    """Get the field name."""
    if self.__field_name__ is None:
      raise MissingVariable('__field_name__', str)
    return self.__field_name__

  def getFieldOwner(self) -> type:
    """Get the field owner."""
    if self.__field_owner__ is None:
      raise MissingVariable('__field_owner__', type)
    return self.__field_owner__

  def _getPositionalArgs(self, **kwargs) -> tuple:
    """Get the positional arguments."""
    if self.__pos_args__ is None:
      return ()
    if isinstance(self.__pos_args__, (tuple, list)):
      instance = maybe(kwargs.get('THIS', self.instance), THIS)
      owner = maybe(kwargs.get('OWNER', self.owner), OWNER)
      desc = maybe(kwargs.get('DESC', self), self)
      posArgs = []
      for arg in self.__pos_args__:
        if arg is THIS:
          posArgs.append(instance)
        elif arg is OWNER:
          posArgs.append(owner)
        else:
          posArgs.append(arg)
      return (*posArgs,)
    raise TypeException('__pos_args__', self.__pos_args__, (tuple, list))

  def _getKeywordArgs(self, **kwargs) -> dict:
    """Get the keyword arguments."""
    if self.__key_args__ is None:
      return {}
    if isinstance(self.__key_args__, dict):
      instance = maybe(kwargs.get('THIS', self.instance), THIS)
      owner = maybe(kwargs.get('OWNER', self.owner), OWNER)
      desc = maybe(kwargs.get('DESC', self), self)
      keyArgs = {}
      for (key, val) in self.__key_args__.items():
        if val is THIS:
          keyArgs[key] = instance
        elif val is OWNER:
          keyArgs[key] = owner
        else:
          keyArgs[key] = val
      return {**keyArgs, }
    raise TypeException('__key_args__', self.__key_args__, dict)

  def getPrivateName(self, ) -> str:
    """Getter-function for the private name of the field."""
    fieldName = self.getFieldName()
    pattern = re.compile(r'(?<!^)(?=[A-Z])')
    return '__%s__' % pattern.sub('_', fieldName).lower()

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  CONSTRUCTORS   # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __init__(self: Any, *args, **kwargs) -> None:
    """Why are we still here?"""
    object.__init__(self, )  # Removed args and kwargs
    self.__pos_args__ = (*args,)
    self.__key_args__ = {**kwargs, }

  def __init_subclass__(cls, *args, **kwargs) -> None:
    """Just to suffer?"""
    object.__init_subclass__()  # Removed args and kwargs

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  PYTHON API   # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __set_name__(self, owner: type, name: str, **kwargs) -> None:
    """
    Provides the instance with awareness of the class, if any, owning it
    and by what name it does. This method is invoked by '__build_class__'
    which is implemented entirely in C.
    """
    self.__field_name__ = name
    self.__field_owner__ = owner

  def __get__(self, instance: Any, owner: type, **kwargs) -> Any:
    """
    After updating the instance and owner, this method returns 'self' if
    accessed from the class (instance is None) or calls the
    '__instance_get__' method.
    """
    self.updateContext(instance, owner, **kwargs)
    if instance is None:
      return self
    try:
      value = self.__instance_get__(**kwargs)
    except Exception as exception:
      raise exception
    else:
      if value is DELETED:
        ownerName = self.getFieldOwner().__name__
        fieldName = self.getFieldName()
        attributeError = attributeErrorFactory(ownerName, fieldName)
        raise attributeError
      return value

  def __set__(self, instance: Any, value: Any, **kwargs) -> None:
    """
    After updating the instance and owner, this method calls the
    '__instance_set__' method.
    """
    self.updateContext(instance, **kwargs)
    return self.__instance_set__(value, **kwargs)

  def __delete__(self, instance: Any, **kwargs) -> None:
    """
    After updating the instance and owner, this method calls the
    '__instance_delete__' method.
    """
    self.updateContext(instance, **kwargs)
    return self.__instance_delete__(**kwargs)

  #  Presentation
  def _getNames(self) -> Names:
    """
    Getter function for class name, owner name and field name. Convenience
    function for '__str__' and '__repr__' methods.
    """
    clsName = type(self).__name__
    modName = type(self).__module__
    if self.owner is None:
      ownerName = ''
      fieldName = ''
    else:
      ownerName = self.owner.__name__
      fieldName = self.__field_name__
    return modName, clsName, ownerName, fieldName

  def __str__(self, ) -> str:
    """
    String representation of the instance. This method should return a
    human-readable string.
    """
    modName, clsName, ownerName, fieldName = self._getNames()
    if ownerName and fieldName:
      infoSpec = """'%s.%s' object at '%s.%s'"""
    else:
      infoSpec = """'%s.%s' object%s%s"""
    return monoSpace(infoSpec % (modName, clsName, ownerName, fieldName))

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  DOMAIN SPECIFIC  # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def updateContext(self, *args, **kwargs) -> None:
    """Updates the current instance."""
    if len(args) == 1:  # __set__ and __delete__ methods
      self.__current_instance__ = args[0]
      self.__current_owner__ = type(args[0])
    elif len(args) == 2:
      ins, own = args
      if not isinstance(ins, own) and ins is not None:
        raise TypeException('instance', ins, own)
      self.__current_instance__ = ins  # even if None
      self.__current_owner__ = own
    else:
      print('__________________________________________________________')
      print("""updateContext problem! Received:""")
      for arg in args:
        print('type: %s, value: %s' % (type(arg), str(arg),))
      print('¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨')
      raise NotImplementedError('precise argument exception not ready')

  def __instance_get__(self, **kwargs, ) -> Any:
    """Subclasses may implement this method to specify the object to be
    returned from __get__. Since this method accepts only the self-argument,
    it should use the current instance and owner as appropriate to
    determine the return-object. By default, this method returns 'self'. """
    return self

  def __instance_set__(self, val: Any, oldVal: Any = None, **kwargs) -> None:
    """Subclasses may implement this method to enable setting of the
    descriptor. By default, this method raises 'ReadOnlyError'. """
    try:
      oldVal = self.__instance_get__()
    except Exception as exception:
      oldVal = exception
    raise ReadOnlyError(self.instance, self, oldVal, val)

  def __instance_delete__(self, oldVal: Any = None, **kwargs) -> None:
    """Subclasses may implement this method to enable deletion of the
    descriptor. By default, this method raises 'ProtectedError'. """
    try:
      oldVal = self.__instance_get__()
    except Exception as exception:
      oldVal = exception
    raise ProtectedError(self.instance, self, oldVal)

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  UTILITIES  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  @classmethod
  def parseKwargs(cls, *args, **kwargs) -> tuple[Any, dict]:
    """
    Parses the keyword arguments for value matching keys and types given
    in positional arguments. The return value is a tuple of the value
    found and the remaining keyword arguments. If no value is found,
    the returned tuple will be 'None' and all the keyword arguments
    received. If no types are given, the type of the value is ignored.
    """
    typeArgs, keys = [], []
    for arg in args:
      if isinstance(arg, type):
        typeArgs.append(arg)
        continue
      if isinstance(arg, str):
        keys.append(arg)
        continue
    if not keys:
      return None, {**kwargs, }
    typeArgs = typeArgs or [object, ]
    if complex in typeArgs:
      if float not in typeArgs:
        typeArgs.append(float)
      if int not in typeArgs:
        typeArgs.append(int)
    elif float in typeArgs:
      if int not in typeArgs:
        typeArgs.append(int)
    for key in keys:
      if key in kwargs:
        value = kwargs[key]
        for type_ in typeArgs:
          if isinstance(value, type_):
            del kwargs[key]
            return value, {**kwargs, }
        else:
          raise TypeException(key, value, *typeArgs, )
    else:
      return None, {**kwargs, }



"""
ReservedNameHook protects reserved names from being overridden.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ...waitaminute import ReservedName

from . import AbstractHook, ReservedNames

try:
  from typing import TYPE_CHECKING
except ImportError:  # pragma: no cover
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any


class ReservedNameHook(AbstractHook):
  """
  ReservedNameHook prevents redefinition of names that are reserved for
  use by Python or the metaclass system. These names are typically
  populated automatically during class construction and must not be
  reassigned by user code.

  ## Protected Names

  The following names are treated as reserved:

  - `__dict__`              – Internal attribute dictionary
  - `__weakref__`           – Weak reference support slot
  - `__module__`            – Name of the module defining the class
  - `__annotations__`       – Type hint storage
  - `__match_args__`        – Structural pattern matching
  - `__doc__`               – Docstring
  - `__name__`              – Class name
  - `__qualname__`          – Fully qualified class name
  - `__firstlineno__`       – Source line for class definition
  - `__static_attributes__` – Internal metadata used by the metaclass

  ## Behavior

  During namespace population (i.e., when a class body assigns names),
  any attempt to override a name in the reserved list will raise a
  ReservedName exception — *but only if* the name already exists in the
  namespace. This prevents accidental clobbering of values while still
  allowing deferred initialization.

  ## Usage
  To use NameHook, simply declare it in your namespace class:

  class Space(AbstractNamespace):  # Must inherit from AbstractNamespace
    #  Custom namespace class inheriting from AbstractNamespace
    reservedNameHook = ReservedNameHook()  # Register the hook
"""

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Public variables
  reservedNames = ReservedNames()

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  DOMAIN SPECIFIC  # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def setItemHook(self, key: str, value: Any, oldValue: Any) -> bool:
    """
    The setItemHook method is called when an item is set in the
    namespace.
    """
    if key in self.reservedNames:
      if key in self.space:
        raise ReservedName(key)
    return False



"""
Nuthin much happenin here!
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

import builtins

from worktoy.waitaminute import MetaclassException, TypeException

oldBuild = builtins.__build_class__


def _resolveMetaclass(func, name, *args, **kwargs) -> type:
  """
  Finds the metaclass in the given arguments.
  """
  mcls = type
  if 'metaclass' in kwargs:
    mcls = kwargs['metaclass']
  elif args:
    if isinstance(args[0], type):
      mcls = type(args[0])
  if isinstance(mcls, type):
    return mcls
  raise TypeException('metaclass', mcls, type)


def _resolveBases(func, name, *args, **kwargs) -> tuple[type, ...]:
  """
  Finds the bases in the given arguments.
  """
  return args


class _InitSub(object):
  """
  A chill object that does not raise any:
  'TypeError: Some.__init_subclass__() takes no keyword arguments'
  """

  def __init__(self, *args, **kwargs) -> None:
    """
    Why are we still here?
    """
    object.__init__(self)

  def __init_subclass__(cls, **kwargs) -> None:
    """
    Just to suffer?
    """
    object.__init_subclass__()


def newBuild(func, name, *args, **kwargs):
  """A new build function that does nothing."""
  mcls = _resolveMetaclass(func, name, *args, **kwargs)
  bases = _resolveBases(func, name, *args, **kwargs)
  try:
    cls = oldBuild(func, name, *args, **kwargs)
  except TypeError as typeError:
    if '__init_subclass__() takes no keyword arguments' in str(typeError):
      return newBuild(func, name, _InitSub, *args, **kwargs)
    if 'metaclass conflict' in str(typeError):
      raise MetaclassException(mcls, name, *bases)
    raise typeError
  else:
    return cls
  finally:
    pass


builtins.__build_class__ = newBuild
